# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from libcst.codemod import CodemodTest
from libcst.codemod.commands.add_pyre_directive import AddPyreUnsafeCommand


class TestAddPyreUnsafeCommand(CodemodTest):

    TRANSFORM = AddPyreUnsafeCommand

    def test_add_to_file(self) -> None:
        before = """
            def baz() -> List[Foo]:
                pass
        """
        after = """
            # pyre-unsafe
            def baz() -> List[Foo]:
                pass
        """
        self.assertCodemod(before, after)

    def test_add_to_file_with_pyre_unsafe(self) -> None:
        """
        We shouldn't be adding pyre-unsafe to a file that already has it.
        """
        before = """
            # pyre-unsafe
            def baz() -> List[Foo]:
                pass
        """
        after = """
            # pyre-unsafe
            def baz() -> List[Foo]:
                pass
        """
        self.assertCodemod(before, after)

    def test_add_to_file_with_pyre_unsafe_after(self) -> None:
        """
        We shouldn't be adding pyre-unsafe to a file that already has it.
        """
        before = """
            # THIS IS A COMMENT!
            # pyre-unsafe
            def baz() -> List[Foo]:
                pass
        """
        after = """
            # THIS IS A COMMENT!
            # pyre-unsafe
            def baz() -> List[Foo]:
                pass
        """
        self.assertCodemod(before, after)

    def test_add_to_file_with_pyre_unsafe_before(self) -> None:
        """
        We shouldn't be adding pyre-unsafe to a file that already has it.
        """
        before = """
            # pyre-unsafe
            # THIS IS A COMMENT!
            def baz() -> List[Foo]:
                pass
        """
        after = """
            # pyre-unsafe
            # THIS IS A COMMENT!
            def baz() -> List[Foo]:
                pass
        """
        self.assertCodemod(before, after)

    def test_empty_file(self) -> None:
        """
        If a file is empty, we should still add it.
        """
        before = ""
        after = "# pyre-unsafe"
        self.assertCodemod(before, after)

    def test_add_to_file_with_comment(self) -> None:
        """
        We should add pyre-unsafe after the last comment at the top of a file.
        """
        before = """
            # YO I'M A COMMENT


            def baz() -> List[Foo]:
                pass
        """
        after = """
            # YO I'M A COMMENT
            # pyre-unsafe


            def baz() -> List[Foo]:
                pass
        """
        self.assertCodemod(before, after)

    def test_add_to_file_with_import(self) -> None:
        """
        Tests that adding to a file with an import works properly.
        """
        before = """
            from typing import List

            def baz() -> List[Foo]:
                pass
        """
        after = """
            # pyre-unsafe
            from typing import List

            def baz() -> List[Foo]:
                pass
        """
        self.assertCodemod(before, after)



import contextlib
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator
from unittest import TestCase

from libcst import BaseExpression, Call, Name, matchers as m
from libcst.codemod import (
    CodemodContext,
    VisitorBasedCodemodCommand,
    gather_files,
    parallel_exec_transform_with_prettyprint,
)
from libcst.codemod.visitors import AddImportsVisitor


class PrintToPPrintCommand(VisitorBasedCodemodCommand):
    def leave_Call(self, original_node: Call, updated_node: Call) -> BaseExpression:
        if m.matches(updated_node, m.Call(func=m.Name("print"))):
            AddImportsVisitor.add_needed_import(
                self.context,
                "pprint",
                "pprint",
            )
            return updated_node.with_changes(func=Name("pprint"))
        return super().leave_Call(original_node, updated_node)


@contextlib.contextmanager
def temp_workspace() -> Generator[Path, None, None]:
    cwd = os.getcwd()
    with TemporaryDirectory() as temp_dir:
        try:
            ws = Path(temp_dir).resolve()
            os.chdir(ws)
            yield ws
        finally:
            os.chdir(cwd)


class ToolE2ETest(TestCase):
    def test_leaky_codemod(self) -> None:
        with temp_workspace() as tmp:
            # File to trigger codemod
            example: Path = tmp / "example.py"
            example.write_text("""print("Hello")""")
            # File that should not be modified
            other = tmp / "other.py"
            other.touch()

            # Run command
            command_instance = PrintToPPrintCommand(CodemodContext())
            files = gather_files(".")
            result = parallel_exec_transform_with_prettyprint(
                command_instance,
                files,
                format_code=False,
                hide_progress=True,
            )

            # Check results
            self.assertEqual(2, result.successes)
            self.assertEqual(0, result.skips)
            self.assertEqual(0, result.failures)
            # Expect example.py to be modified
            self.assertIn(
                "from pprint import pprint",
                example.read_text(),
                "import missing in example.py",
            )
            # Expect other.py to NOT be modified
            self.assertNotIn(
                "from pprint import pprint",
                other.read_text(),
                "import found in other.py",
            )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from textwrap import dedent

from libcst import parse_module
from libcst.testing.utils import UnitTest
from libcst.tool import dump


class PrettyPrintNodesTest(UnitTest):
    def test_full_tree(self) -> None:
        module = r"""
            Module(
              body=[
                FunctionDef(
                  name=Name(
                    value='foo',
                    lpar=[],
                    rpar=[],
                  ),
                  params=Parameters(
                    params=[
                      Param(
                        name=Name(
                          value='a',
                          lpar=[],
                          rpar=[],
                        ),
                        annotation=Annotation(
                          annotation=Name(
                            value='str',
                            lpar=[],
                            rpar=[],
                          ),
                          whitespace_before_indicator=SimpleWhitespace(
                            value='',
                          ),
                          whitespace_after_indicator=SimpleWhitespace(
                            value=' ',
                          ),
                        ),
                        equal=MaybeSentinel.DEFAULT,
                        default=None,
                        comma=MaybeSentinel.DEFAULT,
                        star='',
                        whitespace_after_star=SimpleWhitespace(
                          value='',
                        ),
                        whitespace_after_param=SimpleWhitespace(
                          value='',
                        ),
                      ),
                    ],
                    star_arg=MaybeSentinel.DEFAULT,
                    kwonly_params=[],
                    star_kwarg=None,
                    posonly_params=[],
                    posonly_ind=MaybeSentinel.DEFAULT,
                  ),
                  body=IndentedBlock(
                    body=[
                      SimpleStatementLine(
                        body=[
                          Pass(
                            semicolon=Semicolon(
                              whitespace_before=SimpleWhitespace(
                                value=' ',
                              ),
                              whitespace_after=SimpleWhitespace(
                                value=' ',
                              ),
                            ),
                          ),
                          Pass(
                            semicolon=MaybeSentinel.DEFAULT,
                          ),
                        ],
                        leading_lines=[],
                        trailing_whitespace=TrailingWhitespace(
                          whitespace=SimpleWhitespace(
                            value='',
                          ),
                          comment=None,
                          newline=Newline(
                            value=None,
                          ),
                        ),
                      ),
                      SimpleStatementLine(
                        body=[
                          Return(
                            value=None,
                            whitespace_after_return=SimpleWhitespace(
                              value='',
                            ),
                            semicolon=MaybeSentinel.DEFAULT,
                          ),
                        ],
                        leading_lines=[],
                        trailing_whitespace=TrailingWhitespace(
                          whitespace=SimpleWhitespace(
                            value='',
                          ),
                          comment=None,
                          newline=Newline(
                            value=None,
                          ),
                        ),
                      ),
                    ],
                    header=TrailingWhitespace(
                      whitespace=SimpleWhitespace(
                        value='',
                      ),
                      comment=None,
                      newline=Newline(
                        value=None,
                      ),
                    ),
                    indent=None,
                    footer=[],
                  ),
                  decorators=[],
                  returns=Annotation(
                    annotation=Name(
                      value='None',
                      lpar=[],
                      rpar=[],
                    ),
                    whitespace_before_indicator=SimpleWhitespace(
                      value=' ',
                    ),
                    whitespace_after_indicator=SimpleWhitespace(
                      value=' ',
                    ),
                  ),
                  asynchronous=None,
                  leading_lines=[],
                  lines_after_decorators=[],
                  whitespace_after_def=SimpleWhitespace(
                    value=' ',
                  ),
                  whitespace_after_name=SimpleWhitespace(
                    value='',
                  ),
                  whitespace_before_params=SimpleWhitespace(
                    value='',
                  ),
                  whitespace_before_colon=SimpleWhitespace(
                    value='',
                  ),
                ),
              ],
              header=[],
              footer=[],
              encoding='utf-8',
              default_indent='    ',
              default_newline='\n',
              has_trailing_newline=True,
            )
        """

        # Compare against a known string representation, as unmangled from
        # python indent.
        self.assertEqual(
            dedent(module[1:])[:-1],
            dump(
                parse_module("def foo(a: str) -> None:\n    pass ; pass\n    return\n"),
                show_whitespace=True,
                show_defaults=True,
                show_syntax=True,
            ),
        )

    def test_hidden_whitespace(self) -> None:
        module = r"""
            Module(
              body=[
                FunctionDef(
                  name=Name(
                    value='foo',
                    lpar=[],
                    rpar=[],
                  ),
                  params=Parameters(
                    params=[
                      Param(
                        name=Name(
                          value='a',
                          lpar=[],
                          rpar=[],
                        ),
                        annotation=Annotation(
                          annotation=Name(
                            value='str',
                            lpar=[],
                            rpar=[],
                          ),
                        ),
                        equal=MaybeSentinel.DEFAULT,
                        default=None,
                        comma=MaybeSentinel.DEFAULT,
                        star='',
                      ),
                    ],
                    star_arg=MaybeSentinel.DEFAULT,
                    kwonly_params=[],
                    star_kwarg=None,
                    posonly_params=[],
                    posonly_ind=MaybeSentinel.DEFAULT,
                  ),
                  body=IndentedBlock(
                    body=[
                      SimpleStatementLine(
                        body=[
                          Pass(
                            semicolon=Semicolon(),
                          ),
                          Pass(
                            semicolon=MaybeSentinel.DEFAULT,
                          ),
                        ],
                      ),
                      SimpleStatementLine(
                        body=[
                          Return(
                            value=None,
                            semicolon=MaybeSentinel.DEFAULT,
                          ),
                        ],
                      ),
                    ],
                  ),
                  decorators=[],
                  returns=Annotation(
                    annotation=Name(
                      value='None',
                      lpar=[],
                      rpar=[],
                    ),
                  ),
                  asynchronous=None,
                ),
              ],
              encoding='utf-8',
              default_indent='    ',
              default_newline='\n',
              has_trailing_newline=True,
            )
        """

        # Compare against a known string representation, as unmangled from
        # python indent.
        self.assertEqual(
            dedent(module[1:])[:-1],
            dump(
                parse_module("def foo(a: str) -> None:\n    pass ; pass\n    return\n"),
                show_whitespace=False,
                show_defaults=True,
                show_syntax=True,
            ),
        )

    def test_hidden_defaults(self) -> None:
        module = r"""
            Module(
              body=[
                FunctionDef(
                  name=Name(
                    value='foo',
                  ),
                  params=Parameters(
                    params=[
                      Param(
                        name=Name(
                          value='a',
                        ),
                        annotation=Annotation(
                          annotation=Name(
                            value='str',
                          ),
                          whitespace_before_indicator=SimpleWhitespace(
                            value='',
                          ),
                        ),
                        star='',
                      ),
                    ],
                  ),
                  body=IndentedBlock(
                    body=[
                      SimpleStatementLine(
                        body=[
                          Pass(
                            semicolon=Semicolon(
                              whitespace_before=SimpleWhitespace(
                                value=' ',
                              ),
                              whitespace_after=SimpleWhitespace(
                                value=' ',
                              ),
                            ),
                          ),
                          Pass(),
                        ],
                      ),
                      SimpleStatementLine(
                        body=[
                          Return(
                            whitespace_after_return=SimpleWhitespace(
                              value='',
                            ),
                          ),
                        ],
                      ),
                    ],
                  ),
                  returns=Annotation(
                    annotation=Name(
                      value='None',
                    ),
                    whitespace_before_indicator=SimpleWhitespace(
                      value=' ',
                    ),
                  ),
                ),
              ],
            )
        """

        # Compare against a known string representation, as unmangled from
        # python indent.
        self.assertEqual(
            dedent(module[1:])[:-1],
            dump(
                parse_module("def foo(a: str) -> None:\n    pass ; pass\n    return\n"),
                show_whitespace=True,
                show_defaults=False,
                show_syntax=True,
            ),
        )

    def test_hidden_whitespace_and_defaults(self) -> None:
        module = r"""
            Module(
              body=[
                FunctionDef(
                  name=Name(
                    value='foo',
                  ),
                  params=Parameters(
                    params=[
                      Param(
                        name=Name(
                          value='a',
                        ),
                        annotation=Annotation(
                          annotation=Name(
                            value='str',
                          ),
                        ),
                        star='',
                      ),
                    ],
                  ),
                  body=IndentedBlock(
                    body=[
                      SimpleStatementLine(
                        body=[
                          Pass(
                            semicolon=Semicolon(),
                          ),
                          Pass(),
                        ],
                      ),
                      SimpleStatementLine(
                        body=[
                          Return(),
                        ],
                      ),
                    ],
                  ),
                  returns=Annotation(
                    annotation=Name(
                      value='None',
                    ),
                  ),
                ),
              ],
            )
        """

        # Compare against a known string representation, as unmangled from
        # python indent.
        self.assertEqual(
            dedent(module[1:])[:-1],
            dump(
                parse_module("def foo(a: str) -> None:\n    pass ; pass\n    return\n"),
                show_whitespace=False,
                show_defaults=False,
                show_syntax=True,
            ),
        )

    def test_hidden_syntax(self) -> None:
        module = r"""
            Module(
              body=[
                FunctionDef(
                  name=Name(
                    value='foo',
                    lpar=[],
                    rpar=[],
                  ),
                  params=Parameters(
                    params=[
                      Param(
                        name=Name(
                          value='a',
                          lpar=[],
                          rpar=[],
                        ),
                        annotation=Annotation(
                          annotation=Name(
                            value='str',
                            lpar=[],
                            rpar=[],
                          ),
                          whitespace_before_indicator=SimpleWhitespace(
                            value='',
                          ),
                          whitespace_after_indicator=SimpleWhitespace(
                            value=' ',
                          ),
                        ),
                        default=None,
                        star='',
                        whitespace_after_star=SimpleWhitespace(
                          value='',
                        ),
                        whitespace_after_param=SimpleWhitespace(
                          value='',
                        ),
                      ),
                    ],
                    star_arg=MaybeSentinel.DEFAULT,
                    kwonly_params=[],
                    star_kwarg=None,
                    posonly_params=[],
                    posonly_ind=MaybeSentinel.DEFAULT,
                  ),
                  body=IndentedBlock(
                    body=[
                      SimpleStatementLine(
                        body=[
                          Pass(),
                          Pass(),
                        ],
                        leading_lines=[],
                        trailing_whitespace=TrailingWhitespace(
                          whitespace=SimpleWhitespace(
                            value='',
                          ),
                          comment=None,
                          newline=Newline(
                            value=None,
                          ),
                        ),
                      ),
                      SimpleStatementLine(
                        body=[
                          Return(
                            value=None,
                            whitespace_after_return=SimpleWhitespace(
                              value='',
                            ),
                          ),
                        ],
                        leading_lines=[],
                        trailing_whitespace=TrailingWhitespace(
                          whitespace=SimpleWhitespace(
                            value='',
                          ),
                          comment=None,
                          newline=Newline(
                            value=None,
                          ),
                        ),
                      ),
                    ],
                    header=TrailingWhitespace(
                      whitespace=SimpleWhitespace(
                        value='',
                      ),
                      comment=None,
                      newline=Newline(
                        value=None,
                      ),
                    ),
                    indent=None,
                    footer=[],
                  ),
                  decorators=[],
                  returns=Annotation(
                    annotation=Name(
                      value='None',
                      lpar=[],
                      rpar=[],
                    ),
                    whitespace_before_indicator=SimpleWhitespace(
                      value=' ',
                    ),
                    whitespace_after_indicator=SimpleWhitespace(
                      value=' ',
                    ),
                  ),
                  asynchronous=None,
                  leading_lines=[],
                  lines_after_decorators=[],
                  whitespace_after_def=SimpleWhitespace(
                    value=' ',
                  ),
                  whitespace_after_name=SimpleWhitespace(
                    value='',
                  ),
                  whitespace_before_params=SimpleWhitespace(
                    value='',
                  ),
                  whitespace_before_colon=SimpleWhitespace(
                    value='',
                  ),
                ),
              ],
              header=[],
              footer=[],
            )
        """

        # Compare against a known string representation, as unmangled from
        # python indent.
        self.assertEqual(
            dedent(module[1:])[:-1],
            dump(
                parse_module("def foo(a: str) -> None:\n    pass ; pass\n    return\n"),
                show_whitespace=True,
                show_defaults=True,
                show_syntax=False,
            ),
        )

    def test_hidden_whitespace_and_syntax(self) -> None:
        module = r"""
            Module(
              body=[
                FunctionDef(
                  name=Name(
                    value='foo',
                    lpar=[],
                    rpar=[],
                  ),
                  params=Parameters(
                    params=[
                      Param(
                        name=Name(
                          value='a',
                          lpar=[],
                          rpar=[],
                        ),
                        annotation=Annotation(
                          annotation=Name(
                            value='str',
                            lpar=[],
                            rpar=[],
                          ),
                        ),
                        default=None,
                        star='',
                      ),
                    ],
                    star_arg=MaybeSentinel.DEFAULT,
                    kwonly_params=[],
                    star_kwarg=None,
                    posonly_params=[],
                    posonly_ind=MaybeSentinel.DEFAULT,
                  ),
                  body=IndentedBlock(
                    body=[
                      SimpleStatementLine(
                        body=[
                          Pass(),
                          Pass(),
                        ],
                      ),
                      SimpleStatementLine(
                        body=[
                          Return(
                            value=None,
                          ),
                        ],
                      ),
                    ],
                  ),
                  decorators=[],
                  returns=Annotation(
                    annotation=Name(
                      value='None',
                      lpar=[],
                      rpar=[],
                    ),
                  ),
                  asynchronous=None,
                ),
              ],
            )
        """

        # Compare against a known string representation, as unmangled from
        # python indent.
        self.assertEqual(
            dedent(module[1:])[:-1],
            dump(
                parse_module("def foo(a: str) -> None:\n    pass ; pass\n    return\n"),
                show_whitespace=False,
                show_defaults=True,
                show_syntax=False,
            ),
        )

    def test_hidden_defaults_and_syntax(self) -> None:
        module = r"""
            Module(
              body=[
                FunctionDef(
                  name=Name(
                    value='foo',
                  ),
                  params=Parameters(
                    params=[
                      Param(
                        name=Name(
                          value='a',
                        ),
                        annotation=Annotation(
                          annotation=Name(
                            value='str',
                          ),
                          whitespace_before_indicator=SimpleWhitespace(
                            value='',
                          ),
                        ),
                        star='',
                      ),
                    ],
                  ),
                  body=IndentedBlock(
                    body=[
                      SimpleStatementLine(
                        body=[
                          Pass(),
                          Pass(),
                        ],
                      ),
                      SimpleStatementLine(
                        body=[
                          Return(
                            whitespace_after_return=SimpleWhitespace(
                              value='',
                            ),
                          ),
                        ],
                      ),
                    ],
                  ),
                  returns=Annotation(
                    annotation=Name(
                      value='None',
                    ),
                    whitespace_before_indicator=SimpleWhitespace(
                      value=' ',
                    ),
                  ),
                ),
              ],
            )
        """

        # Compare against a known string representation, as unmangled from
        # python indent.
        self.assertEqual(
            dedent(module[1:])[:-1],
            dump(
                parse_module("def foo(a: str) -> None:\n    pass ; pass\n    return\n"),
                show_whitespace=True,
                show_defaults=False,
                show_syntax=False,
            ),
        )

    def test_hidden_whitespace_and_defaults_and_syntax(self) -> None:
        module = r"""
            Module(
              body=[
                FunctionDef(
                  name=Name(
                    value='foo',
                  ),
                  params=Parameters(
                    params=[
                      Param(
                        name=Name(
                          value='a',
                        ),
                        annotation=Annotation(
                          annotation=Name(
                            value='str',
                          ),
                        ),
                        star='',
                      ),
                    ],
                  ),
                  body=IndentedBlock(
                    body=[
                      SimpleStatementLine(
                        body=[
                          Pass(),
                          Pass(),
                        ],
                      ),
                      SimpleStatementLine(
                        body=[
                          Return(),
                        ],
                      ),
                    ],
                  ),
                  returns=Annotation(
                    annotation=Name(
                      value='None',
                    ),
                  ),
                ),
              ],
            )
        """

        # Compare against a known string representation, as unmangled from
        # python indent.
        self.assertEqual(
            dedent(module[1:])[:-1],
            dump(
                parse_module("def foo(a: str) -> None:\n    pass ; pass\n    return\n"),
                show_whitespace=False,
                show_defaults=False,
                show_syntax=False,
            ),
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import replace
from typing import Generator

from libcst import MetadataDependent, MetadataWrapper, Module
from libcst.codemod._context import CodemodContext


class Codemod(MetadataDependent, ABC):
    """
    Abstract base class that all codemods must subclass from. Classes wishing
    to perform arbitrary, non-visitor-based mutations on a tree should subclass
    from this class directly. Classes wishing to perform visitor-based mutation
    should instead subclass from :class:`~libcst.codemod.ContextAwareTransformer`.

    Note that a :class:`~libcst.codemod.Codemod` is a subclass of
    :class:`~libcst.MetadataDependent`, meaning that you can declare metadata
    dependencies with the :attr:`~libcst.MetadataDependent.METADATA_DEPENDENCIES`
    class property and while you are executing a transform you can call
    :meth:`~libcst.MetadataDependent.get_metadata` to retrieve
    the resolved metadata.
    """

    def __init__(self, context: CodemodContext) -> None:
        MetadataDependent.__init__(self)
        self.context: CodemodContext = context

    def should_allow_multiple_passes(self) -> bool:
        """
        Override this and return ``True`` to allow your transform to be called
        repeatedly until the tree doesn't change between passes. By default,
        this is off, and should suffice for most transforms.
        """
        return False

    def warn(self, warning: str) -> None:
        """
        Emit a warning that is displayed to the user who has invoked this codemod.
        """
        self.context.warnings.append(warning)

    @property
    def module(self) -> Module:
        """
        Reference to the currently-traversed module. Note that this is only available
        during the execution of a codemod. The module reference is particularly
        handy if you want to use :meth:`libcst.Module.code_for_node` or
        :attr:`libcst.Module.config_for_parsing` and don't wish to track a reference
        to the top-level module manually.
        """
        module = self.context.module
        if module is None:
            raise Exception(
                f"Attempted access of {self.__class__.__name__}.module outside of "
                + "transform_module()."
            )
        return module

    @abstractmethod
    def transform_module_impl(self, tree: Module) -> Module:
        """
        Override this with your transform. You should take in the tree, optionally
        mutate it and then return the mutated version. The module reference and all
        calculated metadata are available for the lifetime of this function.
        """
        ...

    @contextmanager
    def _handle_metadata_reference(
        self, module: Module
    ) -> Generator[Module, None, None]:
        oldwrapper = self.context.wrapper
        metadata_manager = self.context.metadata_manager
        filename = self.context.filename
        if metadata_manager and filename:
            # We can look up full-repo metadata for this codemod!
            cache = metadata_manager.get_cache_for_path(filename)
            wrapper = MetadataWrapper(module, cache=cache)
        else:
            # We are missing either the repo manager or the current path,
            # which can happen when we are codemodding from stdin or when
            # an upstream dependency manually instantiates us.
            wrapper = MetadataWrapper(module)

        with self.resolve(wrapper):
            self.context = replace(self.context, wrapper=wrapper)
            try:
                yield wrapper.module
            finally:
                self.context = replace(self.context, wrapper=oldwrapper)

    def transform_module(self, tree: Module) -> Module:
        """
        Transform entrypoint which handles multi-pass logic and metadata calculation
        for you. This is the method that you should call if you wish to invoke a
        codemod directly. This is the method that is called by
        :func:`~libcst.codemod.transform_module`.
        """

        if not self.should_allow_multiple_passes():
            with self._handle_metadata_reference(tree) as tree_with_metadata:
                return self.transform_module_impl(tree_with_metadata)

        # We allow multiple passes, so we execute 1+ passes until there are
        # no more changes.
        previous: Module = tree
        while True:
            with self._handle_metadata_reference(tree) as tree_with_metadata:
                tree = self.transform_module_impl(tree_with_metadata)
            if tree.deep_equals(previous):
                break
            previous = tree
        return tree



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from typing import Dict, Sequence, Union

import libcst as cst
import libcst.matchers as m
import libcst.metadata as meta
from libcst.testing.utils import UnitTest


class MatchersReplaceTest(UnitTest):
    def test_replace_sentinel(self) -> None:
        def _swap_bools(
            node: cst.CSTNode,
            extraction: Dict[str, Union[cst.CSTNode, Sequence[cst.CSTNode]]],
        ) -> cst.CSTNode:
            return cst.Name(
                "True" if cst.ensure_type(node, cst.Name).value == "False" else "False"
            )

        # Verify behavior when provided a sentinel
        replaced = m.replace(
            cst.RemovalSentinel.REMOVE, m.Name("True") | m.Name("False"), _swap_bools
        )
        self.assertEqual(replaced, cst.RemovalSentinel.REMOVE)
        replaced = m.replace(
            cst.MaybeSentinel.DEFAULT, m.Name("True") | m.Name("False"), _swap_bools
        )
        self.assertEqual(replaced, cst.MaybeSentinel.DEFAULT)

    def test_replace_noop(self) -> None:
        def _swap_bools(
            node: cst.CSTNode,
            extraction: Dict[str, Union[cst.CSTNode, Sequence[cst.CSTNode]]],
        ) -> cst.CSTNode:
            return cst.Name(
                "True" if cst.ensure_type(node, cst.Name).value == "False" else "False"
            )

        # Verify behavior when there's nothing to replace.
        original = cst.parse_module("foo: int = 5\ndef bar() -> str:\n    return 's'\n")
        replaced = cst.ensure_type(
            m.replace(original, m.Name("True") | m.Name("False"), _swap_bools),
            cst.Module,
        )
        # Should be identical tree contents
        self.assertTrue(original.deep_equals(replaced))
        # However, should be a new tree by identity
        self.assertNotEqual(id(original), id(replaced))

    def test_replace_simple(self) -> None:
        # Verify behavior when there's a static node as a replacement
        original = cst.parse_module(
            "foo: bool = True\ndef bar() -> bool:\n    return False\n"
        )
        replaced = cst.ensure_type(
            m.replace(original, m.Name("True") | m.Name("False"), cst.Name("boolean")),
            cst.Module,
        ).code
        self.assertEqual(
            replaced, "foo: bool = boolean\ndef bar() -> bool:\n    return boolean\n"
        )

    def test_replace_simple_sentinel(self) -> None:
        # Verify behavior when there's a sentinel as a replacement
        original = cst.parse_module(
            "def bar(x: int, y: int) -> bool:\n    return False\n"
        )
        replaced = cst.ensure_type(
            m.replace(original, m.Param(), cst.RemoveFromParent()), cst.Module
        ).code
        self.assertEqual(replaced, "def bar() -> bool:\n    return False\n")

    def test_replace_actual(self) -> None:
        def _swap_bools(
            node: cst.CSTNode,
            extraction: Dict[str, Union[cst.CSTNode, Sequence[cst.CSTNode]]],
        ) -> cst.CSTNode:
            return cst.Name(
                "True" if cst.ensure_type(node, cst.Name).value == "False" else "False"
            )

        # Verify behavior when there's lots to replace.
        original = cst.parse_module(
            "foo: bool = True\ndef bar() -> bool:\n    return False\n"
        )
        replaced = cst.ensure_type(
            m.replace(original, m.Name("True") | m.Name("False"), _swap_bools),
            cst.Module,
        ).code
        self.assertEqual(
            replaced, "foo: bool = False\ndef bar() -> bool:\n    return True\n"
        )

    def test_replace_add_one(self) -> None:
        def _add_one(
            node: cst.CSTNode,
            extraction: Dict[str, Union[cst.CSTNode, Sequence[cst.CSTNode]]],
        ) -> cst.CSTNode:
            return cst.Integer(str(int(cst.ensure_type(node, cst.Integer).value) + 1))

        # Verify slightly more complex transform behavior.
        original = cst.parse_module("foo: int = 36\ndef bar() -> int:\n    return 41\n")
        replaced = cst.ensure_type(
            m.replace(original, m.Integer(), _add_one), cst.Module
        ).code
        self.assertEqual(replaced, "foo: int = 37\ndef bar() -> int:\n    return 42\n")

    def test_replace_add_one_to_foo_args(self) -> None:
        def _add_one_to_arg(
            node: cst.CSTNode,
            extraction: Dict[str, Union[cst.CSTNode, Sequence[cst.CSTNode]]],
        ) -> cst.CSTNode:
            return node.deep_replace(
                # This can be either a node or a sequence, pyre doesn't know.
                cst.ensure_type(extraction["arg"], cst.CSTNode),
                # Grab the arg and add one to its value.
                cst.Integer(
                    str(int(cst.ensure_type(extraction["arg"], cst.Integer).value) + 1)
                ),
            )

        # Verify way more complex transform behavior.
        original = cst.parse_module(
            "foo: int = 37\ndef bar(baz: int) -> int:\n    return baz\n\nbiz: int = bar(41)\n"
        )
        replaced = cst.ensure_type(
            m.replace(
                original,
                m.Call(
                    func=m.Name("bar"),
                    args=[m.Arg(m.SaveMatchedNode(m.Integer(), "arg"))],
                ),
                _add_one_to_arg,
            ),
            cst.Module,
        ).code
        self.assertEqual(
            replaced,
            "foo: int = 37\ndef bar(baz: int) -> int:\n    return baz\n\nbiz: int = bar(42)\n",
        )

    def test_replace_sequence_extract(self) -> None:
        def _reverse_params(
            node: cst.CSTNode,
            extraction: Dict[str, Union[cst.CSTNode, Sequence[cst.CSTNode]]],
        ) -> cst.CSTNode:
            return cst.ensure_type(node, cst.FunctionDef).with_changes(
                # pyre-ignore We know "params" is a Sequence[Parameters] but asserting that
                # to pyre is difficult.
                params=cst.Parameters(params=list(reversed(extraction["params"])))
            )

        # Verify that we can still extract sequences with replace.
        original = cst.parse_module(
            "def bar(baz: int, foo: int, ) -> int:\n    return baz + foo\n"
        )
        replaced = cst.ensure_type(
            m.replace(
                original,
                m.FunctionDef(
                    params=m.Parameters(
                        params=m.SaveMatchedNode([m.ZeroOrMore(m.Param())], "params")
                    )
                ),
                _reverse_params,
            ),
            cst.Module,
        ).code
        self.assertEqual(
            replaced, "def bar(foo: int, baz: int, ) -> int:\n    return baz + foo\n"
        )

    def test_replace_metadata(self) -> None:
        def _rename_foo(
            node: cst.CSTNode,
            extraction: Dict[str, Union[cst.CSTNode, Sequence[cst.CSTNode]]],
        ) -> cst.CSTNode:
            return cst.ensure_type(node, cst.Name).with_changes(value="replaced")

        original = cst.parse_module(
            "foo: int = 37\ndef bar(foo: int) -> int:\n    return foo\n\nbiz: int = bar(42)\n"
        )
        wrapper = cst.MetadataWrapper(original)
        replaced = cst.ensure_type(
            m.replace(
                wrapper,
                m.Name(
                    metadata=m.MatchMetadataIfTrue(
                        meta.QualifiedNameProvider,
                        lambda qualnames: any(n.name == "foo" for n in qualnames),
                    )
                ),
                _rename_foo,
            ),
            cst.Module,
        ).code
        self.assertEqual(
            replaced,
            "replaced: int = 37\ndef bar(foo: int) -> int:\n    return foo\n\nbiz: int = bar(42)\n",
        )

    def test_replace_metadata_on_transform(self) -> None:
        def _rename_foo(
            node: cst.CSTNode,
            extraction: Dict[str, Union[cst.CSTNode, Sequence[cst.CSTNode]]],
        ) -> cst.CSTNode:
            return cst.ensure_type(node, cst.Name).with_changes(value="replaced")

        original = cst.parse_module(
            "foo: int = 37\ndef bar(foo: int) -> int:\n    return foo\n\nbiz: int = bar(42)\n"
        )
        wrapper = cst.MetadataWrapper(original)

        class TestTransformer(m.MatcherDecoratableTransformer):
            METADATA_DEPENDENCIES: Sequence[meta.ProviderT] = (
                meta.QualifiedNameProvider,
            )

            def leave_Module(
                self, original_node: cst.Module, updated_node: cst.Module
            ) -> cst.Module:
                # Somewhat contrived scenario to test codepaths.
                return cst.ensure_type(
                    self.replace(
                        original_node,
                        m.Name(
                            metadata=m.MatchMetadataIfTrue(
                                meta.QualifiedNameProvider,
                                lambda qualnames: any(
                                    n.name == "foo" for n in qualnames
                                ),
                            )
                        ),
                        _rename_foo,
                    ),
                    cst.Module,
                )

        replaced = cst.ensure_type(wrapper.visit(TestTransformer()), cst.Module).code
        self.assertEqual(
            replaced,
            "replaced: int = 37\ndef bar(foo: int) -> int:\n    return foo\n\nbiz: int = bar(42)\n",
        )

    def test_replace_updated_node_changes(self) -> None:
        def _replace_nested(
            node: cst.CSTNode,
            extraction: Dict[str, Union[cst.CSTNode, Sequence[cst.CSTNode]]],
        ) -> cst.CSTNode:
            return cst.ensure_type(node, cst.Call).with_changes(
                args=[
                    cst.Arg(
                        cst.Name(
                            value=cst.ensure_type(
                                cst.ensure_type(extraction["inner"], cst.Call).func,
                                cst.Name,
                            ).value
                            + "_immediate"
                        )
                    )
                ]
            )

        original = cst.parse_module(
            "def foo(val: int) -> int:\n    return val\nbar = foo\nbaz = foo\nbiz = foo\nfoo(bar(baz(biz(5))))\n"
        )
        replaced = cst.ensure_type(
            m.replace(
                original,
                m.Call(args=[m.Arg(m.SaveMatchedNode(m.Call(), "inner"))]),
                _replace_nested,
            ),
            cst.Module,
        ).code
        self.assertEqual(
            replaced,
            "def foo(val: int) -> int:\n    return val\nbar = foo\nbaz = foo\nbiz = foo\nfoo(bar_immediate)\n",
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# Usage:
#
# python -m libcst.codegen.generate --help
# python -m libcst.codegen.generate visitors

import argparse
import os
import os.path
import shutil
import subprocess
import sys
from typing import List

import libcst as cst
from libcst import ensure_type, parse_module
from libcst.codegen.transforms import (
    DoubleQuoteForwardRefsTransformer,
    SimplifyUnionsTransformer,
)


def format_file(fname: str) -> None:
    with open(os.devnull, "w") as devnull:
        subprocess.check_call(["isort", "-q", fname], stdout=devnull, stderr=devnull)
        subprocess.check_call(["black", fname], stdout=devnull, stderr=devnull)


def clean_generated_code(code: str) -> str:
    """
    Generalized sanity clean-up for all codegen so we can fix issues such as
    Union[SingleType]. The transforms found here are strictly for form and
    do not affect functionality.
    """
    module = parse_module(code)
    module = ensure_type(module.visit(SimplifyUnionsTransformer()), cst.Module)
    module = ensure_type(module.visit(DoubleQuoteForwardRefsTransformer()), cst.Module)
    return module.code


def codegen_visitors() -> None:
    # First, back up the original file, since we have a nasty bootstrap problem.
    # We're in a situation where we want to import libcst in order to get the
    # valid nodes for visitors, but doing so means that we depend on ourselves.
    # So, this attempts to keep the repo in a working state for as many operations
    # as possible.
    base = os.path.abspath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "../")
    )
    visitors_file = os.path.join(base, "_typed_visitor.py")
    shutil.copyfile(visitors_file, f"{visitors_file}.bak")

    try:
        # Now that we backed up the file, lets codegen a new version.
        # We import now, because this script does work on import.
        import libcst.codegen.gen_visitor_functions as visitor_codegen

        new_code = clean_generated_code("\n".join(visitor_codegen.generated_code))
        with open(visitors_file, "w") as fp:
            fp.write(new_code)
            fp.close()

        # Now, see if the file we generated causes any import errors
        # by attempting to run codegen again in a new process.
        with open(os.devnull, "w") as devnull:
            subprocess.check_call(
                ["python3", "-m", "libcst.codegen.gen_visitor_functions"],
                cwd=base,
                stdout=devnull,
            )

        # If it worked, lets format the file
        format_file(visitors_file)

        # Since we were successful with importing, we can remove the backup.
        os.remove(f"{visitors_file}.bak")

        # Inform the user
        print(f"Successfully generated a new {visitors_file} file.")
    except Exception:
        # On failure, we put the original file back, and keep the failed version
        # for developers to look at.
        print(
            f"Failed to generated a new {visitors_file} file, failure "
            + f"is saved in {visitors_file}.failed_generate.",
            file=sys.stderr,
        )
        os.rename(visitors_file, f"{visitors_file}.failed_generate")
        os.rename(f"{visitors_file}.bak", visitors_file)

        # Reraise so we can debug
        raise


def codegen_matchers() -> None:
    # Given that matchers isn't in the default import chain, we don't have to
    # worry about generating invalid code that then prevents us from generating
    # again.
    import libcst.codegen.gen_matcher_classes as matcher_codegen

    base = os.path.abspath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "../")
    )
    matchers_file = os.path.join(base, "matchers/__init__.py")
    new_code = clean_generated_code("\n".join(matcher_codegen.generated_code))
    with open(matchers_file, "w") as fp:
        fp.write(new_code)
        fp.close()

    # If it worked, lets format the file
    format_file(matchers_file)

    # Inform the user
    print(f"Successfully generated a new {matchers_file} file.")


def codegen_return_types() -> None:
    # Given that matchers isn't in the default import chain, we don't have to
    # worry about generating invalid code that then prevents us from generating
    # again.
    import libcst.codegen.gen_type_mapping as type_codegen

    base = os.path.abspath(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "../")
    )
    type_mapping_file = os.path.join(base, "matchers/_return_types.py")
    new_code = clean_generated_code("\n".join(type_codegen.generated_code))
    with open(type_mapping_file, "w") as fp:
        fp.write(new_code)
        fp.close()

    # If it worked, lets format the file
    format_file(type_mapping_file)

    # Inform the user
    print(f"Successfully generated a new {type_mapping_file} file.")


def main(cli_args: List[str]) -> int:
    # Parse out arguments, run codegen
    parser = argparse.ArgumentParser(description="Generate code for libcst.")
    parser.add_argument(
        "system",
        metavar="SYSTEM",
        help='System to generate code for. Valid values include: "visitors", "matchers"',
        type=str,
    )
    args = parser.parse_args(cli_args)
    if args.system == "visitors":
        codegen_visitors()
        return 0
    elif args.system == "matchers":
        codegen_matchers()
        return 0
    elif args.system == "return_types":
        codegen_return_types()
        return 0
    else:
        print(f'Invalid system "{args.system}".')
        return 1


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_expression
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class CallTest(CSTNodeTest):
    @data_provider(
        (
            # Simple call
            {
                "node": cst.Call(cst.Name("foo")),
                "code": "foo()",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"), whitespace_before_args=cst.SimpleWhitespace(" ")
                ),
                "code": "foo( )",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Call with attribute dereference
            {
                "node": cst.Call(cst.Attribute(cst.Name("foo"), cst.Name("bar"))),
                "code": "foo.bar()",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Positional arguments render test
            {
                "node": cst.Call(cst.Name("foo"), (cst.Arg(cst.Integer("1")),)),
                "code": "foo(1)",
                "parser": None,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(cst.Integer("1")),
                        cst.Arg(cst.Integer("2")),
                        cst.Arg(cst.Integer("3")),
                    ),
                ),
                "code": "foo(1, 2, 3)",
                "parser": None,
                "expected_position": None,
            },
            # Positional arguments parse test
            {
                "node": cst.Call(cst.Name("foo"), (cst.Arg(value=cst.Integer("1")),)),
                "code": "foo(1)",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(
                            value=cst.Integer("1"),
                            whitespace_after_arg=cst.SimpleWhitespace(" "),
                        ),
                    ),
                    whitespace_after_func=cst.SimpleWhitespace(" "),
                    whitespace_before_args=cst.SimpleWhitespace(" "),
                ),
                "code": "foo ( 1 )",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(
                            value=cst.Integer("1"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                    ),
                    whitespace_after_func=cst.SimpleWhitespace(" "),
                    whitespace_before_args=cst.SimpleWhitespace(" "),
                ),
                "code": "foo ( 1, )",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(
                            value=cst.Integer("1"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            value=cst.Integer("2"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(value=cst.Integer("3")),
                    ),
                ),
                "code": "foo(1, 2, 3)",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Keyword arguments render test
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (cst.Arg(keyword=cst.Name("one"), value=cst.Integer("1")),),
                ),
                "code": "foo(one = 1)",
                "parser": None,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(keyword=cst.Name("one"), value=cst.Integer("1")),
                        cst.Arg(keyword=cst.Name("two"), value=cst.Integer("2")),
                        cst.Arg(keyword=cst.Name("three"), value=cst.Integer("3")),
                    ),
                ),
                "code": "foo(one = 1, two = 2, three = 3)",
                "parser": None,
                "expected_position": None,
            },
            # Keyword arguments parser test
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(
                            keyword=cst.Name("one"),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("1"),
                        ),
                    ),
                ),
                "code": "foo(one = 1)",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(
                            keyword=cst.Name("one"),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("1"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            keyword=cst.Name("two"),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("2"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            keyword=cst.Name("three"),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("3"),
                        ),
                    ),
                ),
                "code": "foo(one = 1, two = 2, three = 3)",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Iterator expansion render test
            {
                "node": cst.Call(
                    cst.Name("foo"), (cst.Arg(star="*", value=cst.Name("one")),)
                ),
                "code": "foo(*one)",
                "parser": None,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(star="*", value=cst.Name("one")),
                        cst.Arg(star="*", value=cst.Name("two")),
                        cst.Arg(star="*", value=cst.Name("three")),
                    ),
                ),
                "code": "foo(*one, *two, *three)",
                "parser": None,
                "expected_position": None,
            },
            # Iterator expansion parser test
            {
                "node": cst.Call(
                    cst.Name("foo"), (cst.Arg(star="*", value=cst.Name("one")),)
                ),
                "code": "foo(*one)",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(
                            star="*",
                            value=cst.Name("one"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="*",
                            value=cst.Name("two"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(star="*", value=cst.Name("three")),
                    ),
                ),
                "code": "foo(*one, *two, *three)",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Dictionary expansion render test
            {
                "node": cst.Call(
                    cst.Name("foo"), (cst.Arg(star="**", value=cst.Name("one")),)
                ),
                "code": "foo(**one)",
                "parser": None,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(star="**", value=cst.Name("one")),
                        cst.Arg(star="**", value=cst.Name("two")),
                        cst.Arg(star="**", value=cst.Name("three")),
                    ),
                ),
                "code": "foo(**one, **two, **three)",
                "parser": None,
                "expected_position": None,
            },
            # Dictionary expansion parser test
            {
                "node": cst.Call(
                    cst.Name("foo"), (cst.Arg(star="**", value=cst.Name("one")),)
                ),
                "code": "foo(**one)",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(
                            star="**",
                            value=cst.Name("one"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="**",
                            value=cst.Name("two"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(star="**", value=cst.Name("three")),
                    ),
                ),
                "code": "foo(**one, **two, **three)",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Complicated mingling rules render test
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(value=cst.Name("pos1")),
                        cst.Arg(star="*", value=cst.Name("list1")),
                        cst.Arg(value=cst.Name("pos2")),
                        cst.Arg(value=cst.Name("pos3")),
                        cst.Arg(star="*", value=cst.Name("list2")),
                        cst.Arg(value=cst.Name("pos4")),
                        cst.Arg(star="*", value=cst.Name("list3")),
                        cst.Arg(keyword=cst.Name("kw1"), value=cst.Integer("1")),
                        cst.Arg(star="*", value=cst.Name("list4")),
                        cst.Arg(keyword=cst.Name("kw2"), value=cst.Integer("2")),
                        cst.Arg(star="*", value=cst.Name("list5")),
                        cst.Arg(keyword=cst.Name("kw3"), value=cst.Integer("3")),
                        cst.Arg(star="**", value=cst.Name("dict1")),
                        cst.Arg(keyword=cst.Name("kw4"), value=cst.Integer("4")),
                        cst.Arg(star="**", value=cst.Name("dict2")),
                    ),
                ),
                "code": "foo(pos1, *list1, pos2, pos3, *list2, pos4, *list3, kw1 = 1, *list4, kw2 = 2, *list5, kw3 = 3, **dict1, kw4 = 4, **dict2)",
                "parser": None,
                "expected_position": None,
            },
            # Complicated mingling rules parser test
            {
                "node": cst.Call(
                    cst.Name("foo"),
                    (
                        cst.Arg(
                            value=cst.Name("pos1"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="*",
                            value=cst.Name("list1"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            value=cst.Name("pos2"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            value=cst.Name("pos3"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="*",
                            value=cst.Name("list2"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            value=cst.Name("pos4"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="*",
                            value=cst.Name("list3"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            keyword=cst.Name("kw1"),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("1"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="*",
                            value=cst.Name("list4"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            keyword=cst.Name("kw2"),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("2"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="*",
                            value=cst.Name("list5"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            keyword=cst.Name("kw3"),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("3"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="**",
                            value=cst.Name("dict1"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            keyword=cst.Name("kw4"),
                            equal=cst.AssignEqual(),
                            value=cst.Integer("4"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(star="**", value=cst.Name("dict2")),
                    ),
                ),
                "code": "foo(pos1, *list1, pos2, pos3, *list2, pos4, *list3, kw1 = 1, *list4, kw2 = 2, *list5, kw3 = 3, **dict1, kw4 = 4, **dict2)",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Test whitespace
            {
                "node": cst.Call(
                    lpar=(cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),),
                    func=cst.Name("foo"),
                    whitespace_after_func=cst.SimpleWhitespace(" "),
                    whitespace_before_args=cst.SimpleWhitespace(" "),
                    args=(
                        cst.Arg(
                            keyword=None,
                            value=cst.Name("pos1"),
                            comma=cst.Comma(
                                whitespace_before=cst.SimpleWhitespace(" "),
                                whitespace_after=cst.SimpleWhitespace("  "),
                            ),
                        ),
                        cst.Arg(
                            star="*",
                            whitespace_after_star=cst.SimpleWhitespace("  "),
                            keyword=None,
                            value=cst.Name("list1"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            keyword=cst.Name("kw1"),
                            equal=cst.AssignEqual(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            value=cst.Integer("1"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="**",
                            keyword=None,
                            whitespace_after_star=cst.SimpleWhitespace(" "),
                            value=cst.Name("dict1"),
                            whitespace_after_arg=cst.SimpleWhitespace(" "),
                        ),
                    ),
                    rpar=(cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),),
                ),
                "code": "( foo ( pos1 ,  *  list1, kw1=1, ** dict1 ) )",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 2), (1, 43)),
            },
            # Test args
            {
                "node": cst.Arg(
                    star="*",
                    whitespace_after_star=cst.SimpleWhitespace("  "),
                    keyword=None,
                    value=cst.Name("list1"),
                    comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                ),
                "code": "*  list1, ",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 8)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            # Basic expression parenthesizing tests.
            {
                "get_node": lambda: cst.Call(
                    func=cst.Name("foo"), lpar=(cst.LeftParen(),)
                ),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": lambda: cst.Call(
                    func=cst.Name("foo"), rpar=(cst.RightParen(),)
                ),
                "expected_re": "right paren without left paren",
            },
            # Test that we handle keyword stuff correctly.
            {
                "get_node": lambda: cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(
                            equal=cst.AssignEqual(), value=cst.SimpleString("'baz'")
                        ),
                    ),
                ),
                "expected_re": "Must have a keyword when specifying an AssignEqual",
            },
            # Test that we separate *, ** and keyword args correctly
            {
                "get_node": lambda: cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(
                            star="*",
                            keyword=cst.Name("bar"),
                            value=cst.SimpleString("'baz'"),
                        ),
                    ),
                ),
                "expected_re": "Cannot specify a star and a keyword together",
            },
            # Test for expected star inputs only
            {
                "get_node": lambda: cst.Call(
                    func=cst.Name("foo"),
                    # pyre-ignore: Ignore type on 'star' since we're testing behavior
                    # when somebody isn't using a type checker.
                    args=(cst.Arg(star="***", value=cst.SimpleString("'baz'")),),
                ),
                "expected_re": r"Must specify either '', '\*' or '\*\*' for star",
            },
            # Test ordering exceptions
            {
                "get_node": lambda: cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(star="**", value=cst.Name("bar")),
                        cst.Arg(star="*", value=cst.Name("baz")),
                    ),
                ),
                "expected_re": "Cannot have iterable argument unpacking after keyword argument unpacking",
            },
            {
                "get_node": lambda: cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(star="**", value=cst.Name("bar")),
                        cst.Arg(value=cst.Name("baz")),
                    ),
                ),
                "expected_re": "Cannot have positional argument after keyword argument unpacking",
            },
            {
                "get_node": lambda: cst.Call(
                    func=cst.Name("foo"),
                    args=(
                        cst.Arg(
                            keyword=cst.Name("arg"), value=cst.SimpleString("'baz'")
                        ),
                        cst.Arg(value=cst.SimpleString("'bar'")),
                    ),
                ),
                "expected_re": "Cannot have positional argument after keyword argument",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class SmallStatementTest(CSTNodeTest):
    @data_provider(
        (
            {"node": cst.Pass(), "code": "pass"},
            {"node": cst.Pass(semicolon=cst.Semicolon()), "code": "pass;"},
            {
                "node": cst.Pass(
                    semicolon=cst.Semicolon(
                        whitespace_before=cst.SimpleWhitespace("  "),
                        whitespace_after=cst.SimpleWhitespace("    "),
                    )
                ),
                "code": "pass  ;    ",
                "expected_position": CodeRange((1, 0), (1, 4)),
            },
            {"node": cst.Continue(), "code": "continue"},
            {"node": cst.Continue(semicolon=cst.Semicolon()), "code": "continue;"},
            {
                "node": cst.Continue(
                    semicolon=cst.Semicolon(
                        whitespace_before=cst.SimpleWhitespace("  "),
                        whitespace_after=cst.SimpleWhitespace("    "),
                    )
                ),
                "code": "continue  ;    ",
                "expected_position": CodeRange((1, 0), (1, 8)),
            },
            {"node": cst.Break(), "code": "break"},
            {"node": cst.Break(semicolon=cst.Semicolon()), "code": "break;"},
            {
                "node": cst.Break(
                    semicolon=cst.Semicolon(
                        whitespace_before=cst.SimpleWhitespace("  "),
                        whitespace_after=cst.SimpleWhitespace("    "),
                    )
                ),
                "code": "break  ;    ",
                "expected_position": CodeRange((1, 0), (1, 5)),
            },
            {
                "node": cst.Expr(
                    cst.BinaryOperation(cst.Name("x"), cst.Add(), cst.Name("y"))
                ),
                "code": "x + y",
            },
            {
                "node": cst.Expr(
                    cst.BinaryOperation(cst.Name("x"), cst.Add(), cst.Name("y")),
                    semicolon=cst.Semicolon(),
                ),
                "code": "x + y;",
            },
            {
                "node": cst.Expr(
                    cst.BinaryOperation(cst.Name("x"), cst.Add(), cst.Name("y")),
                    semicolon=cst.Semicolon(
                        whitespace_before=cst.SimpleWhitespace("  "),
                        whitespace_after=cst.SimpleWhitespace("    "),
                    ),
                ),
                "code": "x + y  ;    ",
                "expected_position": CodeRange((1, 0), (1, 5)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from unittest import expectedFailure

import libcst as cst
import libcst.matchers as m
from libcst.codemod import Codemod, CodemodContext, CodemodTest, SkipFile


class SimpleCodemod(Codemod):
    def __init__(self, context: CodemodContext, *, skip: bool) -> None:
        super().__init__(context)
        self.skip = skip

    def transform_module_impl(self, tree: cst.Module) -> cst.Module:
        if self.skip:
            raise SkipFile()
        else:
            return tree


class TestSkipDetection(CodemodTest):

    TRANSFORM = SimpleCodemod

    def test_detect_skip(self) -> None:
        code = """
            def foo() -> None:
                pass

            def bar() -> int:
                return 5
        """

        self.assertCodemod(code, code, skip=False, expected_skip=False)
        self.assertCodemod(code, code, skip=True, expected_skip=True)

    @expectedFailure
    def test_did_not_skip_but_should(self) -> None:
        code = """
            def foo() -> None:
                pass

            def bar() -> int:
                return 5
        """

        self.assertCodemod(code, code, skip=False, expected_skip=True)

    @expectedFailure
    def test_skipped_but_should_not(self) -> None:
        code = """
            def foo() -> None:
                pass

            def bar() -> int:
                return 5
        """

        self.assertCodemod(code, code, skip=True, expected_skip=False)


class IncrementCodemod(Codemod):
    def __init__(self, context: CodemodContext, *, iterations: int) -> None:
        super().__init__(context)
        self.iterations = iterations

    def should_allow_multiple_passes(self) -> bool:
        return True

    def transform_module_impl(self, tree: cst.Module) -> cst.Module:
        if self.iterations == 0:
            return tree
        self.iterations -= 1

        return cst.ensure_type(
            m.replace(
                tree,
                m.Integer(),
                lambda node, _: node.with_changes(value=str(int(node.value) + 1)),
            ),
            cst.Module,
        )


class TestMultipass(CodemodTest):

    TRANSFORM = IncrementCodemod

    def test_multi_iterations(self) -> None:
        before = """
            x = 5
        """
        after = """
            x = 10
        """

        self.assertCodemod(before, after, iterations=5)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from textwrap import dedent
from typing import Sequence, Set, Tuple

import libcst as cst
import libcst.matchers as m
import libcst.metadata as meta
from libcst.matchers import matches
from libcst.testing.utils import UnitTest


class MatchersMetadataTest(UnitTest):
    def _make_fixture(
        self, code: str
    ) -> Tuple[cst.BaseExpression, meta.MetadataWrapper]:
        module = cst.parse_module(dedent(code))
        wrapper = cst.MetadataWrapper(module)
        return (
            cst.ensure_type(
                cst.ensure_type(wrapper.module.body[0], cst.SimpleStatementLine).body[
                    0
                ],
                cst.Expr,
            ).value,
            wrapper,
        )

    def _make_coderange(
        self, start: Tuple[int, int], end: Tuple[int, int]
    ) -> meta.CodeRange:
        return meta.CodeRange(
            start=meta.CodePosition(line=start[0], column=start[1]),
            end=meta.CodePosition(line=end[0], column=end[1]),
        )

    def test_simple_matcher_true(self) -> None:
        # Match on a simple node based on the type and the position.
        node, wrapper = self._make_fixture("foo")
        self.assertTrue(
            matches(
                node,
                m.Name(
                    value="foo",
                    metadata=m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 0), (1, 3)),
                    ),
                ),
                metadata_resolver=wrapper,
            )
        )
        # Match on any binary expression where the two children are in exact spots.
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(
            matches(
                node,
                m.BinaryOperation(
                    left=m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 0), (1, 1)),
                    ),
                    right=m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 4), (1, 5)),
                    ),
                ),
                metadata_resolver=wrapper,
            )
        )

    def test_simple_matcher_false(self) -> None:
        # Fail to match on a simple node based on the type and the position.
        node, wrapper = self._make_fixture("foo")
        self.assertFalse(
            matches(
                node,
                m.Name(
                    value="foo",
                    metadata=m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((2, 0), (2, 3)),
                    ),
                ),
                metadata_resolver=wrapper,
            )
        )
        # Fail to match on any binary expression where the two children are in exact spots.
        node, wrapper = self._make_fixture("foo + bar")
        self.assertFalse(
            matches(
                node,
                m.BinaryOperation(
                    left=m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 0), (1, 1)),
                    ),
                    right=m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 4), (1, 5)),
                    ),
                ),
                metadata_resolver=wrapper,
            )
        )

    def test_predicate_logic(self) -> None:
        # Verify that we can or things together.
        matcher = m.BinaryOperation(
            left=m.OneOf(
                m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 1))
                ),
                m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 2))
                ),
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("12 + 3")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("123 + 4")
        self.assertFalse(matches(node, matcher, metadata_resolver=wrapper))

        # Verify that we can and things together
        matcher = m.BinaryOperation(
            left=m.AllOf(
                m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 1))
                ),
                m.MatchMetadata(
                    meta.ExpressionContextProvider, meta.ExpressionContext.LOAD
                ),
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("ab + cd")
        self.assertFalse(matches(node, matcher, metadata_resolver=wrapper))

        # Verify that we can not things
        matcher = m.BinaryOperation(
            left=m.DoesNotMatch(
                m.MatchMetadata(
                    meta.ExpressionContextProvider, meta.ExpressionContext.STORE
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))

    def test_predicate_logic_operators(self) -> None:
        # Verify that we can or things together.
        matcher = m.BinaryOperation(
            left=(
                m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 1))
                )
                | m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 2))
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("12 + 3")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("123 + 4")
        self.assertFalse(matches(node, matcher, metadata_resolver=wrapper))

        # Verify that we can and things together
        matcher = m.BinaryOperation(
            left=(
                m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 1))
                )
                & m.MatchMetadata(
                    meta.ExpressionContextProvider, meta.ExpressionContext.LOAD
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("ab + cd")
        self.assertFalse(matches(node, matcher, metadata_resolver=wrapper))

        # Verify that we can not things
        matcher = m.BinaryOperation(
            left=(
                ~(
                    m.MatchMetadata(
                        meta.ExpressionContextProvider, meta.ExpressionContext.STORE
                    )
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))

    def test_predicate_logic_on_attributes(self) -> None:
        # Verify that we can or things together.
        matcher = m.BinaryOperation(
            left=m.Name(
                metadata=m.OneOf(
                    m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 0), (1, 1)),
                    ),
                    m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 0), (1, 2)),
                    ),
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        matcher = m.BinaryOperation(
            left=m.Integer(
                metadata=m.OneOf(
                    m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 0), (1, 1)),
                    ),
                    m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 0), (1, 2)),
                    ),
                )
            )
        )
        node, wrapper = self._make_fixture("12 + 3")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("123 + 4")
        self.assertFalse(matches(node, matcher, metadata_resolver=wrapper))

        # Verify that we can and things together
        matcher = m.BinaryOperation(
            left=m.Name(
                metadata=m.AllOf(
                    m.MatchMetadata(
                        meta.PositionProvider,
                        self._make_coderange((1, 0), (1, 1)),
                    ),
                    m.MatchMetadata(
                        meta.ExpressionContextProvider, meta.ExpressionContext.LOAD
                    ),
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("ab + cd")
        self.assertFalse(matches(node, matcher, metadata_resolver=wrapper))

        # Verify that we can not things
        matcher = m.BinaryOperation(
            left=m.Name(
                metadata=m.DoesNotMatch(
                    m.MatchMetadata(
                        meta.ExpressionContextProvider, meta.ExpressionContext.STORE
                    )
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))

    def test_predicate_logic_operators_on_attributes(self) -> None:
        # Verify that we can or things together.
        matcher = m.BinaryOperation(
            left=m.Name(
                metadata=m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 1))
                )
                | m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 2))
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        matcher = m.BinaryOperation(
            left=m.Integer(
                metadata=m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 1))
                )
                | m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 2))
                )
            )
        )
        node, wrapper = self._make_fixture("12 + 3")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("123 + 4")
        self.assertFalse(matches(node, matcher, metadata_resolver=wrapper))

        # Verify that we can and things together
        matcher = m.BinaryOperation(
            left=m.Name(
                metadata=m.MatchMetadata(
                    meta.PositionProvider, self._make_coderange((1, 0), (1, 1))
                )
                & m.MatchMetadata(
                    meta.ExpressionContextProvider, meta.ExpressionContext.LOAD
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))
        node, wrapper = self._make_fixture("ab + cd")
        self.assertFalse(matches(node, matcher, metadata_resolver=wrapper))

        # Verify that we can not things
        matcher = m.BinaryOperation(
            left=m.Name(
                metadata=~(
                    m.MatchMetadata(
                        meta.ExpressionContextProvider, meta.ExpressionContext.STORE
                    )
                )
            )
        )
        node, wrapper = self._make_fixture("a + b")
        self.assertTrue(matches(node, matcher, metadata_resolver=wrapper))

    def test_lambda_metadata_matcher(self) -> None:
        # Match on qualified name provider
        module = cst.parse_module(
            "from typing import List\n\ndef foo() -> None: pass\n"
        )
        wrapper = cst.MetadataWrapper(module)
        functiondef = cst.ensure_type(wrapper.module.body[1], cst.FunctionDef)

        self.assertTrue(
            matches(
                functiondef,
                m.FunctionDef(
                    name=m.MatchMetadataIfTrue(
                        meta.QualifiedNameProvider,
                        lambda qualnames: any(
                            n.name in {"foo", "bar", "baz"} for n in qualnames
                        ),
                    )
                ),
                metadata_resolver=wrapper,
            )
        )

        self.assertFalse(
            matches(
                functiondef,
                m.FunctionDef(
                    name=m.MatchMetadataIfTrue(
                        meta.QualifiedNameProvider,
                        lambda qualnames: any(
                            n.name in {"bar", "baz"} for n in qualnames
                        ),
                    )
                ),
                metadata_resolver=wrapper,
            )
        )

    def test_lambda_metadata_matcher_with_no_metadata(self) -> None:
        # Match on qualified name provider
        module = cst.parse_module(
            "from typing import List\n\ndef foo() -> None: pass\n"
        )
        functiondef = cst.ensure_type(module.body[1], cst.FunctionDef)

        self.assertFalse(
            matches(
                functiondef,
                m.FunctionDef(
                    name=m.MatchMetadataIfTrue(
                        meta.QualifiedNameProvider,
                        lambda qualnames: any(
                            n.name in {"foo", "bar", "baz"} for n in qualnames
                        ),
                    )
                ),
            )
        )

    def test_lambda_metadata_matcher_operators(self) -> None:
        # Match on qualified name provider
        module = cst.parse_module(
            "from typing import List\n\ndef bar() -> None: pass\n"
        )
        wrapper = cst.MetadataWrapper(module)
        functiondef = cst.ensure_type(wrapper.module.body[1], cst.FunctionDef)

        self.assertTrue(
            matches(
                functiondef,
                m.FunctionDef(
                    name=m.MatchMetadataIfTrue(
                        meta.QualifiedNameProvider,
                        lambda qualnames: any(n.name == "foo" for n in qualnames),
                    )
                    | m.MatchMetadataIfTrue(
                        meta.QualifiedNameProvider,
                        lambda qualnames: any(n.name == "bar" for n in qualnames),
                    )
                ),
                metadata_resolver=wrapper,
            )
        )

        self.assertFalse(
            matches(
                functiondef,
                m.FunctionDef(
                    name=m.MatchMetadataIfTrue(
                        meta.QualifiedNameProvider,
                        lambda qualnames: any(n.name == "foo" for n in qualnames),
                    )
                    & m.MatchMetadataIfTrue(
                        meta.QualifiedNameProvider,
                        lambda qualnames: any(n.name == "bar" for n in qualnames),
                    )
                ),
                metadata_resolver=wrapper,
            )
        )

        self.assertTrue(
            matches(
                functiondef,
                m.FunctionDef(
                    name=(
                        ~m.MatchMetadataIfTrue(
                            meta.QualifiedNameProvider,
                            lambda qualnames: any(n.name == "foo" for n in qualnames),
                        )
                    )
                    & m.MatchMetadataIfTrue(
                        meta.QualifiedNameProvider,
                        lambda qualnames: any(n.name == "bar" for n in qualnames),
                    )
                ),
                metadata_resolver=wrapper,
            )
        )


class MatchersVisitorMetadataTest(UnitTest):
    def _make_fixture(self, code: str) -> cst.MetadataWrapper:
        return cst.MetadataWrapper(cst.parse_module(dedent(code)))

    def test_matches_on_visitors(self) -> None:
        # Set up a simple visitor that has a metadata dependency, try to use it in matchers.
        class TestVisitor(m.MatcherDecoratableVisitor):
            METADATA_DEPENDENCIES: Sequence[meta.ProviderT] = (
                meta.ExpressionContextProvider,
            )

            def __init__(self) -> None:
                super().__init__()
                self.match_names: Set[str] = set()

            def visit_Name(self, node: cst.Name) -> None:
                # Only match name nodes that are being assigned to.
                if self.matches(
                    node,
                    m.Name(
                        metadata=m.MatchMetadata(
                            meta.ExpressionContextProvider, meta.ExpressionContext.STORE
                        )
                    ),
                ):
                    self.match_names.add(node.value)

        module = self._make_fixture(
            """
            a = 1 + 2
            b = 3 + 4 + d + e
            def foo() -> str:
                c = "baz"
                return c
            def bar() -> int:
                return b
            del foo
            del bar
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        self.assertEqual(visitor.match_names, {"a", "b", "c", "foo", "bar"})

    def test_matches_on_transformers(self) -> None:
        # Set up a simple visitor that has a metadata dependency, try to use it in matchers.
        class TestTransformer(m.MatcherDecoratableTransformer):
            METADATA_DEPENDENCIES: Sequence[meta.ProviderT] = (
                meta.ExpressionContextProvider,
            )

            def __init__(self) -> None:
                super().__init__()
                self.match_names: Set[str] = set()

            def visit_Name(self, node: cst.Name) -> None:
                # Only match name nodes that are being assigned to.
                if self.matches(
                    node,
                    m.Name(
                        metadata=m.MatchMetadata(
                            meta.ExpressionContextProvider, meta.ExpressionContext.STORE
                        )
                    ),
                ):
                    self.match_names.add(node.value)

        module = self._make_fixture(
            """
            a = 1 + 2
            b = 3 + 4 + d + e
            def foo() -> str:
                c = "baz"
                return c
            def bar() -> int:
                return b
            del foo
            del bar
        """
        )
        visitor = TestTransformer()
        module.visit(visitor)

        self.assertEqual(visitor.match_names, {"a", "b", "c", "foo", "bar"})

    def test_matches_decorator_on_visitors(self) -> None:
        # Set up a simple visitor that has a metadata dependency, try to use it in matchers.
        class TestVisitor(m.MatcherDecoratableVisitor):
            METADATA_DEPENDENCIES: Sequence[meta.ProviderT] = (
                meta.ExpressionContextProvider,
            )

            def __init__(self) -> None:
                super().__init__()
                self.match_names: Set[str] = set()

            @m.visit(
                m.Name(
                    metadata=m.MatchMetadata(
                        meta.ExpressionContextProvider, meta.ExpressionContext.STORE
                    )
                )
            )
            def _visit_assignments(self, node: cst.Name) -> None:
                # Only match name nodes that are being assigned to.
                self.match_names.add(node.value)

        module = self._make_fixture(
            """
            a = 1 + 2
            b = 3 + 4 + d + e
            def foo() -> str:
                c = "baz"
                return c
            def bar() -> int:
                return b
            del foo
            del bar
        """
        )
        visitor = TestVisitor()
        module.visit(visitor)

        self.assertEqual(visitor.match_names, {"a", "b", "c", "foo", "bar"})

    def test_matches_decorator_on_transformers(self) -> None:
        # Set up a simple visitor that has a metadata dependency, try to use it in matchers.
        class TestTransformer(m.MatcherDecoratableTransformer):
            METADATA_DEPENDENCIES: Sequence[meta.ProviderT] = (
                meta.ExpressionContextProvider,
            )

            def __init__(self) -> None:
                super().__init__()
                self.match_names: Set[str] = set()

            @m.visit(
                m.Name(
                    metadata=m.MatchMetadata(
                        meta.ExpressionContextProvider, meta.ExpressionContext.STORE
                    )
                )
            )
            def _visit_assignments(self, node: cst.Name) -> None:
                # Only match name nodes that are being assigned to.
                self.match_names.add(node.value)

        module = self._make_fixture(
            """
            a = 1 + 2
            b = 3 + 4 + d + e
            def foo() -> str:
                c = "baz"
                return c
            def bar() -> int:
                return b
            del foo
            del bar
        """
        )
        visitor = TestTransformer()
        module.visit(visitor)

        self.assertEqual(visitor.match_names, {"a", "b", "c", "foo", "bar"})



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Collection, List, Set, Union, cast

import libcst as cst
import libcst.matchers as m
from libcst.codemod._context import CodemodContext
from libcst.codemod._visitor import ContextAwareVisitor
from libcst.metadata import MetadataWrapper, QualifiedNameProvider


FUNCS_CONSIDERED_AS_STRING_ANNOTATIONS = {"typing.TypeVar"}


class GatherNamesFromStringAnnotationsVisitor(ContextAwareVisitor):
    """
    Collects all names from string literals used for typing purposes.
    This includes annotations like ``foo: "SomeType"``, and parameters to
    special functions related to typing (currently only `typing.TypeVar`).

    After visiting, a set of all found names will be available on the ``names``
    attribute of this visitor.
    """

    METADATA_DEPENDENCIES = (QualifiedNameProvider,)

    def __init__(
        self,
        context: CodemodContext,
        typing_functions: Collection[str] = FUNCS_CONSIDERED_AS_STRING_ANNOTATIONS,
    ) -> None:
        super().__init__(context)
        self._typing_functions: Collection[str] = typing_functions
        self._annotation_stack: List[cst.CSTNode] = []
        #: The set of names collected from string literals.
        self.names: Set[str] = set()

    def visit_Annotation(self, node: cst.Annotation) -> bool:
        self._annotation_stack.append(node)
        return True

    def leave_Annotation(self, original_node: cst.Annotation) -> None:
        self._annotation_stack.pop()

    def visit_Call(self, node: cst.Call) -> bool:
        qnames = self.get_metadata(QualifiedNameProvider, node)
        if any(qn.name in self._typing_functions for qn in qnames):
            self._annotation_stack.append(node)
            return True
        return False

    def leave_Call(self, original_node: cst.Call) -> None:
        if self._annotation_stack and self._annotation_stack[-1] == original_node:
            self._annotation_stack.pop()

    def visit_ConcatenatedString(self, node: cst.ConcatenatedString) -> bool:
        if self._annotation_stack:
            self.handle_any_string(node)
        return False

    def visit_SimpleString(self, node: cst.SimpleString) -> bool:
        if self._annotation_stack:
            self.handle_any_string(node)
        return False

    def handle_any_string(
        self, node: Union[cst.SimpleString, cst.ConcatenatedString]
    ) -> None:
        value = node.evaluated_value
        if value is None:
            return
        mod = cst.parse_module(value)
        extracted_nodes = m.extractall(
            mod,
            m.Name(
                value=m.SaveMatchedNode(m.DoNotCare(), "name"),
                metadata=m.MatchMetadataIfTrue(
                    cst.metadata.ParentNodeProvider,
                    lambda parent: not isinstance(parent, cst.Attribute),
                ),
            )
            | m.SaveMatchedNode(m.Attribute(), "attribute"),
            metadata_resolver=MetadataWrapper(mod, unsafe_skip_copy=True),
        )
        names = {
            cast(str, values["name"]) for values in extracted_nodes if "name" in values
        } | {
            name
            for values in extracted_nodes
            if "attribute" in values
            for name, _ in cst.metadata.scope_provider._gen_dotted_names(
                cast(cst.Attribute, values["attribute"])
            )
        }
        self.names.update(names)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from itertools import islice
from typing import List

import libcst


def insert_header_comments(node: libcst.Module, comments: List[str]) -> libcst.Module:
    """
    Insert comments after last non-empty line in header. Use this to insert one or more
    comments after any copyright preamble in a :class:`~libcst.Module`. Each comment in
    the list of ``comments`` must start with a ``#`` and will be placed on its own line
    in the appropriate location.
    """
    # Split the lines up into a contiguous comment-containing section and
    # the empty whitespace section that follows
    last_comment_index = -1
    for i, line in enumerate(node.header):
        if line.comment is not None:
            last_comment_index = i

    comment_lines = islice(node.header, last_comment_index + 1)
    empty_lines = islice(node.header, last_comment_index + 1, None)
    inserted_lines = [
        libcst.EmptyLine(comment=libcst.Comment(value=comment)) for comment in comments
    ]
    return node.with_changes(header=(*comment_lines, *inserted_lines, *empty_lines))



# Copyright 2004-2005 Elemental Security, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.
#
# Modifications:
# Copyright David Halter and Contributors
# Modifications are dual-licensed: MIT and PSF.
# 99% of the code is different from pgen2, now.
#
# A fork of `parso.pgen2.grammar_parser`.
# https://github.com/davidhalter/parso/blob/master/parso/pgen2/grammar_parser.py
#
# The following changes were made:
# - Type stubs were directly applied.
# pyre-unsafe

from typing import Generator, List, Optional, Tuple

from libcst._parser.parso.python.token import PythonTokenTypes
from libcst._parser.parso.python.tokenize import tokenize
from libcst._parser.parso.utils import parse_version_string


class NFAArc:
    def __init__(self, next_: "NFAState", nonterminal_or_string: Optional[str]) -> None:
        self.next: NFAState = next_
        self.nonterminal_or_string: Optional[str] = nonterminal_or_string

    def __repr__(self) -> str:
        return "<%s: %s>" % (self.__class__.__name__, self.nonterminal_or_string)


class NFAState:
    def __init__(self, from_rule: str) -> None:
        self.from_rule = from_rule
        self.arcs: List[NFAArc] = []

    def add_arc(
        self, next_: "NFAState", nonterminal_or_string: Optional[str] = None
    ) -> None:
        self.arcs.append(NFAArc(next_, nonterminal_or_string))

    def __repr__(self) -> str:
        return "<%s: from %s>" % (self.__class__.__name__, self.from_rule)


class GrammarParser:
    """
    The parser for Python grammar files.
    """

    def __init__(self, bnf_grammar: str) -> None:
        self._bnf_grammar: str = bnf_grammar
        self.generator = tokenize(bnf_grammar, version_info=parse_version_string("3.6"))
        self._gettoken()  # Initialize lookahead

    def parse(self) -> Generator[Tuple[NFAState, NFAState], None, None]:
        # grammar: (NEWLINE | rule)* ENDMARKER
        while self.type != PythonTokenTypes.ENDMARKER:
            while self.type == PythonTokenTypes.NEWLINE:
                self._gettoken()

            # rule: NAME ':' rhs NEWLINE
            # pyre-ignore Pyre is unhappy with the fact that we haven't put
            # _current_rule_name in the constructor.
            self._current_rule_name = self._expect(PythonTokenTypes.NAME)
            self._expect(PythonTokenTypes.OP, ":")

            a, z = self._parse_rhs()
            self._expect(PythonTokenTypes.NEWLINE)

            yield a, z

    def _parse_rhs(self):
        # rhs: items ('|' items)*
        a, z = self._parse_items()
        if self.value != "|":
            return a, z
        else:
            aa = NFAState(self._current_rule_name)
            zz = NFAState(self._current_rule_name)
            while True:
                # Add the possibility to go into the state of a and come back
                # to finish.
                aa.add_arc(a)
                z.add_arc(zz)
                if self.value != "|":
                    break

                self._gettoken()
                a, z = self._parse_items()
            return aa, zz

    def _parse_items(self):
        # items: item+
        a, b = self._parse_item()
        while (
            self.type
            in (
                PythonTokenTypes.NAME,
                PythonTokenTypes.STRING,
            )
            or self.value in ("(", "[")
        ):
            c, d = self._parse_item()
            # Need to end on the next item.
            b.add_arc(c)
            b = d
        return a, b

    def _parse_item(self):
        # item: '[' rhs ']' | atom ['+' | '*']
        if self.value == "[":
            self._gettoken()
            a, z = self._parse_rhs()
            self._expect(PythonTokenTypes.OP, "]")
            # Make it also possible that there is no token and change the
            # state.
            a.add_arc(z)
            return a, z
        else:
            a, z = self._parse_atom()
            value = self.value
            if value not in ("+", "*"):
                return a, z
            self._gettoken()
            # Make it clear that we can go back to the old state and repeat.
            z.add_arc(a)
            if value == "+":
                return a, z
            else:
                # The end state is the same as the beginning, nothing must
                # change.
                return a, a

    def _parse_atom(self):
        # atom: '(' rhs ')' | NAME | STRING
        if self.value == "(":
            self._gettoken()
            a, z = self._parse_rhs()
            self._expect(PythonTokenTypes.OP, ")")
            return a, z
        elif self.type in (PythonTokenTypes.NAME, PythonTokenTypes.STRING):
            a = NFAState(self._current_rule_name)
            z = NFAState(self._current_rule_name)
            # Make it clear that the state transition requires that value.
            a.add_arc(z, self.value)
            self._gettoken()
            return a, z
        else:
            self._raise_error(
                "expected (...) or NAME or STRING, got %s/%s", self.type, self.value
            )

    def _expect(self, type_, value=None):
        if self.type != type_:
            self._raise_error("expected %s, got %s [%s]", type_, self.type, self.value)
        if value is not None and self.value != value:
            self._raise_error("expected %s, got %s", value, self.value)
        value = self.value
        self._gettoken()
        return value

    def _gettoken(self) -> None:
        tup = next(self.generator)
        self.type, self.value, self.begin, prefix = tup

    def _raise_error(self, msg: str, *args: object) -> None:
        if args:
            try:
                msg = msg % args
            except Exception:
                msg = " ".join([msg] + list(map(str, args)))
        line = self._bnf_grammar.splitlines()[self.begin[0] - 1]
        raise SyntaxError(msg, ("<grammar>", self.begin[0], self.begin[1], line))
